# -*- coding: utf-8 -*-

import logging

from tornado import gen
from tornado.tcpclient import TCPClient
from tornado_discovery.registrant import BaseRegistrant
from tornado_discovery.common import Command, Status

from litepipeline.node.config import CONFIG

LOG = logging.getLogger(__name__)


class Registrant(BaseRegistrant):
    _instance = None

    def __new__(cls, host, port, config, retry_interval = 10, reconnect = True):
        if not cls._instance:
            cls._instance = object.__new__(cls)
            cls._instance.host = host
            cls._instance.port = port
            cls._instance.config = config
            cls._instance.retry_interval = retry_interval
            cls._instance.heartbeat_interval = cls._instance.config.get("heartbeat_interval")
            cls._instance.heartbeat_timeout = cls._instance.config.get("heartbeat_timeout")
            cls._instance.reconnect = reconnect
            cls._instance.tcpclient = TCPClient()
            cls._instance.periodic_heartbeat = None
            cls._instance._stream = None
            cls._instance.heartbeat_data = {}
        return cls._instance

    def __init__(self, host, port, config, retry_interval = 10, reconnect = True):
        pass

    @classmethod
    def instance(cls):
        return cls._instance

    def update_heartbeat_data(self, data = {}):
        self.heartbeat_data.update(data)

    @gen.coroutine
    def heartbeat_service(self):
        try:
            message_data = self.config.to_dict()
            message_data.update(self.heartbeat_data)
            data = {"command": Command.heartbeat, "data": message_data}
            self.send_message(data)
            data = yield self.read_message()
            if data["data"]["status"] == Status.success:
                LOG.info("Client Received Heartbeat Message: %s", data["data"])
            else:
                LOG.error("Client Received Heartbeat Message: %s", data["data"])
        except Exception as e:
            LOG.exception(e)



# -*- coding: utf-8 -*-

import json
import time
import logging

from tornado import web
from tornado import gen

from litepipeline.manager.handlers.base import BaseHandler, BaseSocketHandler
from litepipeline.manager.utils.listener import Connection
from litepipeline.manager.utils.common import Errors
from litepipeline.version import __version__
from litepipeline.manager.config import CONFIG

LOG = logging.getLogger("__name__")


class AboutHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        result = {"message": "LitePipeline manager service"}
        self.write(result)
        self.finish()


class ClusterInfoHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK, "version": __version__}
        try:
            info = {"number_of_nodes": 0, "nodes": []}
            for node in Connection.clients:
                info["nodes"].append(node.info)
                info["number_of_nodes"] += 1
            result["info"] = info
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(json.dumps(result, sort_keys = True))
        self.finish()



# -*- coding: utf-8 -*-

import os
import time
import logging
import shutil
import tarfile

import requests
from tornado import web
from tornado import gen

from litepipeline.manager.handlers.base import BaseHandler, StreamBaseHandler
from litepipeline.manager.models.applications import Applications
from litepipeline.manager.utils.listener import Connection
from litepipeline.manager.utils.common import file_sha1sum, file_md5sum, Errors, splitall
from litepipeline.manager.config import CONFIG

LOG = logging.getLogger("__name__")


class CreateApplicationHandler(StreamBaseHandler):
    @gen.coroutine
    def post(self):
        result = {"result": Errors.OK}
        try:
            name = self.get_form_argument("name", "")
            description = self.get_form_argument("description", "")
            if name and os.path.exists(self.file_path) and os.path.isfile(self.file_path):
                sha1 = file_sha1sum(self.file_path)
                LOG.debug("sha1: %s, %s", sha1, type(sha1))
                app_id = Applications.instance().add(name, sha1, description = description)
                app_path = os.path.join(CONFIG["data_path"], "applications", app_id[:2], app_id[2:4], app_id)
                if os.path.exists(app_path):
                    shutil.rmtree(app_path)
                os.makedirs(app_path)
                shutil.copy2(self.file_path.decode("utf-8"), os.path.join(app_path, "app.tar.gz"))
                os.remove(self.file_path)
                if os.path.exists(os.path.join(app_path, "app")):
                    shutil.rmtree(os.path.join(app_path, "app"))
                t = tarfile.open(os.path.join(app_path, "app.tar.gz"), "r")
                t.extractall(app_path)
                path_parts = splitall(t.getnames()[0])
                tar_root_name = path_parts[1] if path_parts[0] == "." else path_parts[0]
                t.close()
                os.rename(os.path.join(app_path, tar_root_name), os.path.join(app_path, "app"))
                result["app_id"] = app_id
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        LOG.debug("deploy application package use: %ss", time.time() - self.start)
        self.write(result)
        self.finish()


class ListApplicationHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK}
        try:
            offset = int(self.get_argument("offset", "0"))
            limit = int(self.get_argument("limit", "0"))
            LOG.debug("ListApplicationHandler offset: %s, limit: %s", offset, limit)
            result["apps"] = Applications.instance().list(offset = offset, limit = limit)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class DeleteApplicationHandler(BaseHandler):
    @gen.coroutine
    def delete(self):
        result = {"result": Errors.OK}
        try:
            app_id = self.get_argument("app_id", "")
            if app_id:
                success = Applications.instance().delete(app_id)
                if success:
                    app_path = os.path.join(CONFIG["data_path"], "applications", app_id[:2], app_id[2:4], app_id)
                    if os.path.exists(app_path):
                        shutil.rmtree(app_path)
                        LOG.debug("remove directory: %s", app_path) 
                else:
                    Errors.set_result_error("OperationFailed", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class UpdateApplicationHandler(StreamBaseHandler):
    @gen.coroutine
    def put(self):
        result = {"result": Errors.OK}
        try:
            app_id = self.get_form_argument("app_id", "")
            name = self.get_form_argument("name", "")
            description = self.get_form_argument("description", "")
            LOG.debug("UpdateApplicationHandler app_id: %s, name: %s, description: %s", app_id, name, description)
            if app_id and Applications.instance().get(app_id):
                data = {}
                need_update = False
                if name:
                    data["name"] = name
                if description:
                    data["description"] = description
                if os.path.exists(self.file_path) and os.path.isfile(self.file_path):
                    sha1 = file_sha1sum(self.file_path)
                    data["sha1"] = sha1
                    LOG.debug("sha1: %s, %s", sha1, type(sha1))
                    app_path = os.path.join(CONFIG["data_path"], "applications", app_id[:2], app_id[2:4], app_id)
                    if os.path.exists(app_path):
                        shutil.rmtree(app_path)
                    os.makedirs(app_path)
                    shutil.copy2(self.file_path.decode("utf-8"), os.path.join(app_path, "app.tar.gz"))
                    os.remove(self.file_path)
                    if os.path.exists(os.path.join(app_path, "app")):
                        shutil.rmtree(os.path.join(app_path, "app"))
                    t = tarfile.open(os.path.join(app_path, "app.tar.gz"), "r")
                    t.extractall(app_path)
                    tar_root_name = splitall(t.getnames()[0])[0]
                    os.rename(os.path.join(app_path, tar_root_name), os.path.join(app_path, "app"))
                    result["app_id"] = app_id
                    need_update = True
                if data or need_update:
                    success = Applications.instance().update(app_id, data)
                    if not success:
                        Errors.set_result_error("OperationFailed", result)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        LOG.debug("update application package use: %ss", time.time() - self.start)
        self.write(result)
        self.finish()


class InfoApplicationHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK}
        try:
            app_id = self.get_argument("app_id", "")
            if app_id:
                app_info = Applications.instance().get(app_id)
                if app_info:
                    result["app_info"] = app_info
                elif app_info is None:
                    Errors.set_result_error("AppNotExists", result)
                else:
                    Errors.set_result_error("OperationFailed", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class DownloadApplicationHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK}
        try:
            app_id = self.get_argument("app_id", "")
            if app_id:
                app_info = Applications.instance().get(app_id)
                if app_info:
                    app_path = os.path.join(CONFIG["data_path"], "applications", app_id[:2], app_id[2:4], app_id, "app.tar.gz")
                    if os.path.exists(app_path):
                        buf_size = 64 * 1024
                        self.set_header('Content-Type', 'application/octet-stream')
                        self.set_header('Content-Disposition', 'attachment; filename=%s.tar.gz' % app_id)
                        with open(app_path, 'rb') as f:
                            while True:
                                data = f.read(buf_size)
                                if not data:
                                    break
                                self.write(data)
                                self.flush()
                                yield gen.moment
                        self.finish()
                        return
                elif app_info is None:
                    Errors.set_result_error("AppNotExists", result)
                else:
                    Errors.set_result_error("OperationFailed", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()



# -*- coding: utf-8 -*-

import logging

from tinydb import TinyDB
from tinydb.operations import delete

from tornado_discovery.config import BaseConfig

LOG = logging.getLogger(__name__)


class PersistentConfig(BaseConfig):
    def __init__(self, db_path):
        self.db_path = db_path
        self.db = TinyDB(self.db_path)
        self.doc_id = 1
        if len(self.db.all()) == 0:
            self.db.insert({})

    def has_key(self, key):
        result = False
        try:
            result = key in self.db.all()[0]
        except Exception as e:
            LOG.exception(e)
        return result

    def get(self, key):
        result = None
        try:
            result = self.db.all()[0][key]
        except Exception as e:
            LOG.exception(e)
        return result

    def set(self, key, value):
        result = False
        try:
            self.db.update({key: value}, doc_ids = [self.doc_id])
            result = True
        except Exception as e:
            LOG.exception(e)
        return result

    def update(self, key, value):
        return self.set(key, value)

    def delete(self, key):
        result = False
        try:
            self.db.update(delete(key), doc_ids = [self.doc_id])
            result = True
        except Exception as e:
            LOG.exception(e)
        return result

    def to_dict(self):
        result = {}
        try:
            result = self.db.all()[0]
        except Exception as e:
            LOG.exception(e)
        return result

    def from_dict(self, data):
        result = False
        try:
            self.db.update(data, doc_ids = [self.doc_id])
            result = True
        except Exception as e:
            LOG.exception(e)
        return result



# -*- coding: utf-8 -*-

import json
import datetime
import logging
from uuid import uuid4

from litepipeline.manager.db.sqlite_interface import SchedulesTable, NoResultFound
from litepipeline.manager.utils.common import Status, Stage
from litepipeline.manager.config import CONFIG

LOG = logging.getLogger(__name__)


class Schedules(object):
    _instance = None

    def __new__(cls):
        if not cls._instance:
            cls._instance = object.__new__(cls)
            cls._instance.table = SchedulesTable
            engine, session = SchedulesTable.init_engine_and_session()
            cls._instance.table.metadata.create_all(engine)
            cls._instance.session = session(autoflush = False, autocommit = False)
            cls._instance.cache = {}
            cls._instance.load_cache()
        return cls._instance

    @classmethod
    def instance(cls):
        return cls._instance

    def _new_id(self):
        return str(uuid4())

    def load_cache(self):
        schedules = self.list(enable = True)
        for schedule in schedules:
            schedule_id = schedule["schedule_id"]
            self.cache[schedule_id] = schedule

    def add(self, schedule_name, app_id, minute = -1, hour = -1, day_of_month = -1, day_of_week = -1, enable = False, input_data = {}):
        result = False
        schedule_id = self._new_id()
        now = datetime.datetime.now()
        item = {
            "schedule_id": schedule_id,
            "schedule_name": schedule_name,
            "application_id": app_id,
            "create_at": now,
            "update_at": now,
            "minute": minute,
            "hour": hour,
            "day_of_month": day_of_month,
            "day_of_week": day_of_week,
            "input_data": json.dumps(input_data),
            "enable": enable,
        }

        row = self.table()
        row.parse_dict(item)
        try:
            self.session.add(row)
            self.session.commit()
            self.cache[schedule_id] = row.to_dict()
            result = schedule_id
            LOG.debug("add schedule: %s", row)
        except Exception as e:
            LOG.exception(e)
            self.session.rollback()
        return result

    def update(self, schedule_id, data):
        result = False
        try:
            now = datetime.datetime.now()
            if "input_data" in data:
                data["input_data"] = json.dumps(data["input_data"])
            data["update_at"] = now
            self.session.query(self.table).filter_by(schedule_id = schedule_id).update(data)
            self.session.commit()
            if "input_data" in data:
                data["input_data"] = json.loads(data["input_data"])
            data["update_at"] = str(now)
            if schedule_id in self.cache:
                if "enable" in data and not data["enable"]:
                    del self.cache[schedule_id]
                else:
                    self.cache[schedule_id].update(data)
            else:
                schedule = self.get(schedule_id)
                self.cache[schedule_id] = schedule
            result = True
            LOG.debug("update schedule: %s, %s", schedule_id, data)
        except Exception as e:
            LOG.exception(e)
            self.session.rollback()
        return result

    def delete(self, schedule_id):
        result = False
        try:
            row = self.session.query(self.table).filter_by(schedule_id = schedule_id).one()
            self.session.delete(row)
            self.session.commit()
            if schedule_id in self.cache:
                del self.cache[schedule_id]
            result = True
            LOG.debug("delete schedule: %s", row)
        except Exception as e:
            LOG.exception(e)
        return result

    def get(self, schedule_id):
        result = False
        try:
            row = self.session.query(self.table).filter_by(schedule_id = schedule_id).one()
            result = row.to_dict()
        except NoResultFound:
            result = None
        except Exception as e:
            LOG.exception(e)
        return result

    def list(self, offset = 0, limit = 0, enable = None):
        result = []
        try:
            offset = 0 if offset < 0 else offset
            limit = 0 if limit < 0 else limit
            if isinstance(enable, bool):
                if limit:
                    rows = self.session.query(self.table).filter_by(enable = enable).order_by(self.table.update_at.desc()).offset(offset).limit(limit)
                elif offset:
                    rows = self.session.query(self.table).filter_by(enable = enable).order_by(self.table.update_at.desc()).offset(offset)
                else:
                    rows = self.session.query(self.table).filter_by(enable = enable).order_by(self.table.update_at.desc())
            else:
                if limit:
                    rows = self.session.query(self.table).order_by(self.table.update_at.desc()).offset(offset).limit(limit)
                elif offset:
                    rows = self.session.query(self.table).order_by(self.table.update_at.desc()).offset(offset)
                else:
                    rows = self.session.query(self.table).order_by(self.table.update_at.desc())
            for row in rows:
                result.append(row.to_dict())
        except Exception as e:
            LOG.exception(e)
        return result

    def close(self):
        self.session.close()



# -*- coding: utf-8 -*-

import os
import sys
import time
import json
import logging
import shutil
import tarfile
import signal
from pathlib import Path
import threading
from threading import Thread, Lock
from multiprocessing import Process, Queue, Pipe

from tornado import gen, locks
import requests

from litepipeline.node.utils.common import file_sha1sum, splitall
from litepipeline.node.config import CONFIG
from litepipeline.node import logger

LOG = logging.getLogger(__name__)


class Command(object):
    check_app = "check_app"
    exit = "exit"


class TasksCache(object):
    tasks = {}
    tasks_lock = Lock()

    @classmethod
    def set(cls, app_id):
        cls.tasks_lock.acquire()
        if app_id not in cls.tasks:
            cls.tasks[app_id] = False
        cls.tasks_lock.release()

    @classmethod
    def get(cls):
        result = False
        cls.tasks_lock.acquire()
        try:
            for app_id in cls.tasks:
                if not cls.tasks[app_id]:
                    cls.tasks[app_id] = True
                    result = app_id
                    break
        except Exception as e:
            LOG.exception(e)
        cls.tasks_lock.release()
        return result

    @classmethod
    def remove(cls, app_id):
        cls.tasks_lock.acquire()
        del cls.tasks[app_id]
        cls.tasks_lock.release()


class StoppableThread(Thread):
    """
    Thread class with a stop() method. The thread itself has to check
    regularly for the stopped() condition.
    """

    def __init__(self):
        super(StoppableThread, self).__init__()
        self._stop_event = threading.Event()

    def stop(self):
        self._stop_event.set()

    def stopped(self):
        return self._stop_event.isSet()


class WorkerThread(StoppableThread):
    def __init__(self, pid):
        StoppableThread.__init__(self)
        Thread.__init__(self)
        self.pid = pid

    def run(self):
        LOG = logging.getLogger("worker")
        LOG.info("Worker(%03d) start", self.pid)
        try:
            while not self.stopped():
                try:
                    app_id = TasksCache.get()
                    if app_id:
                        url = "http://%s:%s/app/download?app_id=%s" % (CONFIG["manager_http_host"], CONFIG["manager_http_port"], app_id)
                        file_path = os.path.join(CONFIG["data_path"], "tmp", "%s.tar.gz" % app_id)
                        LOG.debug("download: %s", url)
                        r = requests.get(url)
                        if r.status_code == 200:
                            LOG.debug("download[%s] status: %s", app_id, r.status_code)
                            f = open(file_path, 'wb')
                            f.write(r.content)
                            f.close()
                            app_path = os.path.join(CONFIG["data_path"], "applications", app_id[:2], app_id[2:4], app_id)
                            if os.path.exists(app_path):
                                shutil.rmtree(app_path)
                            os.makedirs(app_path)
                            shutil.copy2(file_path, os.path.join(app_path, "app.tar.gz"))
                            os.remove(file_path)
                            if os.path.exists(os.path.join(app_path, "app")):
                                shutil.rmtree(os.path.join(app_path, "app"))
                            t = tarfile.open(os.path.join(app_path, "app.tar.gz"), "r")
                            t.extractall(app_path)
                            path_parts = splitall(t.getnames()[0])
                            tar_root_name = path_parts[1] if path_parts[0] == "." else path_parts[0]
                            t.close()
                            app_config_path = os.path.join(app_path, tar_root_name, "configuration.json")
                            f = open(app_config_path, "r")
                            app_config = json.loads(f.read())
                            f.close()
                            venvs = set()
                            for action in app_config["actions"]:
                                if "env" in action:
                                    venvs.add(action["env"])
                            for venv in list(venvs):
                                venv_tar_path = os.path.join(app_path, tar_root_name, "%s.tar.gz" % venv)
                                venv_path = os.path.join(app_path, tar_root_name, venv)
                                if os.path.exists(venv_path):
                                    shutil.rmtree(venv_path)
                                os.makedirs(venv_path)
                                t = tarfile.open(venv_tar_path, "r")
                                t.extractall(venv_path)
                                t.close()
                            os.rename(os.path.join(app_path, tar_root_name), os.path.join(app_path, "app"))
                            TasksCache.remove(app_id)
                        else:
                            LOG.warning("download[%s] status: %s", app_id, r.status_code)
                    else:
                        time.sleep(0.5)
                except Exception as e:
                    LOG.exception(e)
        except Exception as e:
            LOG.exception(e)
        LOG.info("Worker(%03d) exit", self.pid)


class Manager(Process):
    def __init__(self, pipe_client, worker_num):
        Process.__init__(self)
        self.pipe_client = pipe_client
        self.worker_num = worker_num

    def run(self):
        logger.config_logging(file_name = "apps_manager.log",
                              log_level = "NOSET",
                              dir_name = CONFIG["log_path"],
                              when = "D",
                              interval = 1,
                              max_size = 20,
                              backup_count = 5,
                              console = True)
        LOG = logging.getLogger("manager")

        def sig_handler(sig, frame):
            LOG.warning("sig_handler Caught signal: %s", sig)

        LOG.info("Manager start")
        try:
            signal.signal(signal.SIGTERM, sig_handler)
            signal.signal(signal.SIGINT, sig_handler)

            threads = []
            for i in range(self.worker_num):
                t = WorkerThread(i)
                t.start()
                threads.append(t)

            while True:
                LOG.debug("Manager main loop")
                command, app_id, sha1 = self.pipe_client.recv()
                if command == Command.check_app:
                    ready = False
                    try:
                        LOG.debug("check app, app_id: %s, sha1: %s", app_id, sha1)
                        app_base_path = os.path.join(CONFIG["data_path"], "applications", app_id[:2], app_id[2:4], app_id)
                        app_tar_path = os.path.join(app_base_path, "app.tar.gz")
                        app_path = os.path.join(app_base_path, "app")
                        if os.path.exists(app_tar_path) and os.path.isfile(app_tar_path):
                            if sha1 != file_sha1sum(app_tar_path):
                                # download app.tar.gz && extract app.tar.gz
                                pass
                            else:
                                if not os.path.exists(app_path):
                                    # extract app.tar.gz
                                    pass
                                else:
                                    ready = True
                        # download app.tar.gz && extract app.tar.gz
                        if not ready:
                            TasksCache.set(app_id)
                    except Exception as e:
                        LOG.exception(e)
                    self.pipe_client.send((command, ready))
                elif command == Command.exit:
                    for t in threads:
                        t.stop()
                    break
            for t in threads:
                t.join()
        except Exception as e:
            LOG.exception(e)
        LOG.info("Manager exit")


class ManagerClient(object):
    process_list = []
    process_dict = {}
    write_lock = locks.Lock()
    _instance = None

    def __init__(self, worker_num = 1):
        if ManagerClient._instance is None:
            self.worker_num = worker_num if worker_num > 0 else 1
            LOG.debug("ManagerClient, worker_num: %s", self.worker_num)
            pipe_master, pipe_client = Pipe()
            p = Manager(pipe_client, self.worker_num)
            p.daemon = True
            ManagerClient.process_list.append(p)
            ManagerClient.process_dict["manager"] = [p, pipe_master]
            p.start()
            ManagerClient._instance = self
        else:
            self.worker_num = ManagerClient._instance.worker_num

    @gen.coroutine
    def check_app(self, app_id, sha1):
        result = False
        LOG.debug("start check app, app_id: %s, sha1: %s", app_id, sha1)
        with (yield ManagerClient.write_lock.acquire()):
            LOG.debug("get check app lock, app_id: %s, sha1: %s", app_id, sha1)
            ManagerClient.process_dict["manager"][1].send((Command.check_app, app_id, sha1))
            LOG.debug("send check app, app_id: %s, sha1: %s", app_id, sha1)
            while not ManagerClient.process_dict["manager"][1].poll():
                yield gen.moment
            LOG.debug("recv check app, app_id: %s, sha1: %s", app_id, sha1)
            r = ManagerClient.process_dict["manager"][1].recv()
            LOG.debug("end check app, app_id: %s, sha1: %s, r: %s", app_id, sha1, r)
        if r[1]:
            result = r[1]
        raise gen.Return(result)

    def close(self):
        try:
            LOG.info("close ManagerClient")
            ManagerClient.process_dict["manager"][1].send((Command.exit, None, None))
            for p in ManagerClient.process_list[1:]:
                p.terminate()
            for p in ManagerClient.process_list:
                while p.is_alive():
                    time.sleep(0.5)
                    LOG.debug("sleep 0.5 second")
            LOG.info("All Process Exit!")
        except Exception as e:
            LOG.exception(e)



# -*- coding: utf-8 -*-
'''
Created on 2014-05-07
@summary: A custom logging
@author: YangHaitao
'''

import os
import logging
import logging.handlers

CWD = os.path.split(os.path.realpath(__file__))[0]

LEVELS = {'NOSET': logging.NOTSET,
          'DEBUG': logging.DEBUG,
          'INFO': logging.INFO,
          'WARNING': logging.WARNING,
          'ERROR': logging.ERROR,
          'CRITICAL': logging.CRITICAL}


class ConsoleStreamHandler(logging.StreamHandler):
    # color names to indices
    color_map = {
        'black': 0,
        'red': 1,
        'green': 2,
        'yellow': 3,
        'blue': 4,
        'magenta': 5,
        'cyan': 6,
        'white': 7,
    }
    # levels to (background, foreground, bold/intense)
    level_map = {
        logging.DEBUG: (None, 'blue', False),
        logging.INFO: (None, 'white', False),
        logging.WARNING: (None, 'yellow', False),
        logging.ERROR: (None, 'red', False),
        logging.CRITICAL: ('red', 'white', True),
    }

    csi = '\x1b['
    reset = '\x1b[0m'

    def colorize(self, message, record):
        """
        Colorize a message for a logging event.

        This implementation uses the ``level_map`` class attribute to
        map the LogRecord's level to a colour/intensity setting, which is
        then applied to the whole message.

        :param message: The message to colorize.
        :param record: The ``LogRecord`` for the message.
        """
        if record.levelno in self.level_map:
            bg, fg, bold = self.level_map[record.levelno]
            params = []
            if bg in self.color_map:
                params.append(str(self.color_map[bg] + 40))
            if fg in self.color_map:
                params.append(str(self.color_map[fg] + 30))
            if bold:
                params.append('1')
            if params:
                message = ''.join((self.csi, ';'.join(params), 'm', message, self.reset))
        return message

    def format(self, record):
        """
        Formats a record for output.

        This implementation colorizes the message line, but leaves
        any traceback unolorized.
        """
        message = logging.StreamHandler.format(self, record)
        parts = message.split('\n', 1)
        parts[0] = self.colorize(parts[0], record)
        message = '\n'.join(parts)
        return message

    def emit(self, record):
        try:
            message = self.format(record)
            stream = self.stream
            if bytes and isinstance(message, bytes):
                enc = getattr(stream, 'encoding', 'utf-8')
                if not enc: # just for pyinstaller None bugfix
                    enc = 'utf-8'
                message = message.encode(enc, 'replace')
            stream.write(message)
            stream.write(getattr(self, 'terminator', '\n'))
            self.flush()
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self.handleError(record)


def config_logging(logger_name = "",
                   file_name = "main.log",
                   log_level = "NOSET",
                   dir_name = "logs",
                   day_rotate = False,
                   when = "D",
                   interval = 1,
                   max_size = 50,
                   backup_count = 5,
                   console = True):
    format_log_string = "%(asctime)s %(name)-12s %(levelname)-8s %(message)s"
    format_console_string = "%(name)-12s: %(levelname)-8s %(message)s"
    logs_dir = os.path.join(CWD, dir_name)
    file_dir = os.path.join(logs_dir, file_name)
    # init logs directory
    if os.path.exists(logs_dir) and os.path.isdir(logs_dir):
        pass
    else:
        try:
            os.makedirs(logs_dir)
        except Exception as e:
            pass
    # clear all handlers
    logging.getLogger(logger_name).handlers = []
    # init rotating handler
    if day_rotate == True:
        rotatingFileHandler = logging.handlers.TimedRotatingFileHandler(filename = file_dir,
                                                                        when = when,
                                                                        interval = interval,
                                                                        backupCount = backup_count)
    else:
        rotatingFileHandler = logging.handlers.RotatingFileHandler(filename = file_dir,
                                                                   maxBytes = 1024 * 1024 * max_size,
                                                                   backupCount = backup_count)
    formatter = logging.Formatter(format_log_string)
    rotatingFileHandler.setFormatter(formatter)
    logging.getLogger(logger_name).addHandler(rotatingFileHandler)
    # add a console handler
    if console == True:
        if os.name == 'nt':
            consoleHandler = logging.StreamHandler()
        else:
            consoleHandler = ConsoleStreamHandler()
        # set console log level
        consoleHandler.setLevel(LEVELS[log_level.upper()])
        formatter = logging.Formatter(format_console_string)
        consoleHandler.setFormatter(formatter)
        logging.getLogger(logger_name).addHandler(consoleHandler)
    # set log level
    logger = logging.getLogger(logger_name)
    level = LEVELS[log_level.upper()]
    logger.setLevel(level)


#!/usr/bin/env python3
# -*- coding: UTF-8 -*-

import os
import stat
import argparse
from pathlib import Path

from litepipeline.version import __version__

cwd = os.path.split(os.path.realpath(__file__))[0]
parser = argparse.ArgumentParser(prog = 'liteconfig')

# common arguments
parser.add_argument("-s", "--service", required = True, help = "configuration service (manager, node)")
parser.add_argument("-o", "--output", required = True, help = "configuration file output path")
parser.add_argument("-v", "--version", action = 'version', version = '%(prog)s ' + __version__)

args = parser.parse_args()


def main():
    try:
        service = args.service
        output = args.output
        manager_path = os.path.join(cwd, "../manager")
        node_path = os.path.join(cwd, "../node")
        if service in ("manager", "node"):
            if os.path.exists(output) and os.path.isdir(output):
                output = str(Path(output).resolve())
                log_path = os.path.join(output, "logs")
                data_path = os.path.join(output, "data")
                content = ""
                if service == "manager":
                    manager_config_file = os.path.join(manager_path, "configuration.yml.temp")
                    fp = open(manager_config_file, "r")
                    content = fp.read()
                    fp.close()
                    copy_files = [
                        "install_systemd_service.sh",
                        "uninstall_systemd_service.sh",
                        "litepipeline-manager.service.temp",
                        "manager.sh",
                        "README.md",
                    ]
                    for file_name in copy_files:
                        file_path_source = os.path.join(manager_path, file_name)
                        with open(file_path_source, "r") as fs:
                            file_path_target = os.path.join(output, file_name)
                            with open(file_path_target, "w") as ft:
                                ft.write(fs.read())
                            if file_path_target.endswith(".sh"):
                                os.chmod(
                                    file_path_target,
                                    stat.S_IRUSR | stat.S_IWUSR | stat.S_IEXEC | stat.S_IRGRP | stat.S_IXGRP | stat.S_IROTH | stat.S_IXOTH
                                )
                elif service == "node":
                    node_config_file = os.path.join(node_path, "configuration.yml.temp")
                    fp = open(node_config_file, "r")
                    content = fp.read()
                    fp.close()
                    copy_files = [
                        "install_systemd_service.sh",
                        "uninstall_systemd_service.sh",
                        "litepipeline-node.service.temp",
                        "node.sh",
                        "README.md",
                    ]
                    for file_name in copy_files:
                        file_path_source = os.path.join(node_path, file_name)
                        with open(file_path_source, "r") as fs:
                            file_path_target = os.path.join(output, file_name)
                            with open(file_path_target, "w") as ft:
                                ft.write(fs.read())
                            if file_path_target.endswith(".sh"):
                                os.chmod(
                                    file_path_target,
                                    stat.S_IRUSR | stat.S_IWUSR | stat.S_IEXEC | stat.S_IRGRP | stat.S_IXGRP | stat.S_IROTH | stat.S_IXOTH
                                )
                content = content.replace("log_path_string", log_path)
                content = content.replace("data_path_string", data_path)
                fp = open(os.path.join(output, "configuration.yml"), "w")
                fp.write(content)
                fp.close()
            else:
                print("output(%s) not exists!", output)
        else:
            print("service must by manager or node, not %s", service)
    except Exception as e:
        print(e)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

import datetime
import logging
from uuid import uuid4

from litepipeline.manager.db.sqlite_interface import ApplicationsTable, NoResultFound
from litepipeline.manager.config import CONFIG


LOG = logging.getLogger(__name__)


class Applications(object):
    _instance = None

    def __new__(cls):
        if not cls._instance:
            cls._instance = object.__new__(cls)
            cls._instance.table = ApplicationsTable
            engine, session = ApplicationsTable.init_engine_and_session()
            cls._instance.table.metadata.create_all(engine)
            cls._instance.session = session(autoflush = False, autocommit = False)
        return cls._instance

    @classmethod
    def instance(cls):
        return cls._instance

    def _new_id(self):
        return str(uuid4())

    def add(self, name, sha1, description = ""):
        result = False
        app_id = self._new_id()
        now = datetime.datetime.now()
        item = {
            "application_id": app_id,
            "name": name,
            "create_at": now,
            "update_at": now,
            "sha1": sha1,
            "description": description,
        }

        row = self.table()
        row.parse_dict(item)
        try:
            self.session.add(row)
            self.session.commit()
            result = app_id
            LOG.debug("add application: %s", row)
        except Exception as e:
            LOG.exception(e)
            self.session.rollback()
        return result

    def update(self, app_id, data):
        result = False
        try:
            now = datetime.datetime.now()
            data["update_at"] = now
            self.session.query(self.table).filter_by(application_id = app_id).update(data)
            self.session.commit()
            result = True
            LOG.debug("update application: %s, %s", app_id, data)
        except Exception as e:
            LOG.exception(e)
            self.session.rollback()
        return result

    def delete(self, app_id):
        result = False
        try:
            row = self.session.query(self.table).filter_by(application_id = app_id).one()
            self.session.delete(row)
            self.session.commit()
            result = True
            LOG.debug("delete application: %s", row)
        except Exception as e:
            LOG.exception(e)
        return result

    def get(self, app_id):
        result = False
        try:
            row = self.session.query(self.table).filter_by(application_id = app_id).one()
            result = row.to_dict()
        except NoResultFound:
            result = None
        except Exception as e:
            LOG.exception(e)
        return result

    def list(self, offset = 0, limit = 0):
        result = []
        try:
            offset = 0 if offset < 0 else offset
            limit = 0 if limit < 0 else limit
            if limit:
                rows = self.session.query(self.table).order_by(self.table.update_at.desc()).offset(offset).limit(limit)
            elif offset:
                rows = self.session.query(self.table).order_by(self.table.update_at.desc()).offset(offset)
            else:
                rows = self.session.query(self.table).order_by(self.table.update_at.desc())
            for row in rows:
                result.append(row.to_dict())
        except Exception as e:
            LOG.exception(e)
        return result

    def close(self):
        self.session.close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import signal
import logging
import argparse

import tornado.ioloop
import tornado.httpserver
import tornado.web

from litepipeline.version import __version__
from litepipeline.manager.handlers import info
from litepipeline.manager.handlers import application
from litepipeline.manager.handlers import task
from litepipeline.manager.handlers import schedule
from litepipeline.manager.utils.listener import Connection
from litepipeline.manager.utils.listener import DiscoveryListener
from litepipeline.manager.models.applications import Applications
from litepipeline.manager.models.tasks import Tasks
from litepipeline.manager.models.schedules import Schedules
from litepipeline.manager.utils.scheduler import Scheduler
from litepipeline.manager.utils import common
from litepipeline.manager.config import CONFIG, load_config
from litepipeline.manager import logger

LOG = logging.getLogger(__name__)


class Application(tornado.web.Application):
    def __init__(self):
        handlers = [
            (r"/", info.AboutHandler),
            (r"/cluster/info", info.ClusterInfoHandler),
            (r"/app/create", application.CreateApplicationHandler),
            (r"/app/list", application.ListApplicationHandler),
            (r"/app/delete", application.DeleteApplicationHandler),
            (r"/app/update", application.UpdateApplicationHandler),
            (r"/app/info", application.InfoApplicationHandler),
            (r"/app/download", application.DownloadApplicationHandler),
            (r"/task/create", task.CreateTaskHandler),
            (r"/task/list", task.ListTaskHandler),
            (r"/task/info", task.InfoTaskHandler),
            (r"/task/delete", task.DeleteTaskHandler),
            (r"/task/rerun", task.RunTaskHandler),
            (r"/task/stop", task.StopTaskHandler),
            (r"/action/update", task.UpdateActionHandler),
            (r"/workspace/delete", task.DeleteTaskWorkspaceHandler),
            (r"/workspace/pack", task.PackTaskWorkspaceHandler),
            (r"/workspace/download", task.DownloadTaskWorkspaceHandler),
            (r"/schedule/create", schedule.CreateScheduleHandler),
            (r"/schedule/update", schedule.UpdateScheduleHandler),
            (r"/schedule/list", schedule.ListScheduleHandler),
            (r"/schedule/delete", schedule.DeleteScheduleHandler),
            (r"/schedule/info", schedule.InfoScheduleHandler),
        ]
        settings = dict(debug = False)
        tornado.web.Application.__init__(self, handlers, **settings)


def main():
    parser = argparse.ArgumentParser(prog = 'litemanager')
    parser.add_argument("-c", "--config", required = True, help = "configuration file path")
    parser.add_argument("-v", "--version", action = 'version', version = '%(prog)s ' + __version__)
    args = parser.parse_args()

    if args.config:
        success = load_config(args.config)
        if success:
            common.init_storage()
            logger.config_logging(file_name = "manager.log",
                                  log_level = CONFIG["log_level"],
                                  dir_name = CONFIG["log_path"],
                                  day_rotate = False,
                                  when = "D",
                                  interval = 1,
                                  max_size = 20,
                                  backup_count = 5,
                                  console = True)

            LOG.info("service start")

            try:
                tasks_db = Tasks()
                applications_db = Applications()
                schedules_db = Schedules()
                task_scheduler = Scheduler(CONFIG["scheduler_interval"])
                http_server = tornado.httpserver.HTTPServer(
                    Application(),
                    max_buffer_size = CONFIG["max_buffer_size"],
                    chunk_size = 10 * 1024 * 1024
                )
                http_server.listen(CONFIG["http_port"], address = CONFIG["http_host"])
                # http_server.bind(CONFIG["http_port"], address = CONFIG["http_host"])
                listener = DiscoveryListener(Connection, task_scheduler)
                listener.listen(CONFIG["tcp_port"], CONFIG["tcp_host"])
                common.Servers.HTTP_SERVER = http_server
                common.Servers.DB_SERVERS.append(applications_db)
                common.Servers.DB_SERVERS.append(tasks_db)
                common.Servers.DB_SERVERS.append(schedules_db)
                common.Servers.DB_SERVERS.append(task_scheduler)
                signal.signal(signal.SIGTERM, common.sig_handler)
                signal.signal(signal.SIGINT, common.sig_handler)
                tornado.ioloop.IOLoop.instance().start()
            except Exception as e:
                LOG.exception(e)

            LOG.info("service end")
        else:
            print("failed to load configuration: %s" % args.config)
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

import json
import time
import urllib
import logging

from tornado import web
from tornado import gen

from litepipeline.manager.handlers.base import BaseHandler, BaseSocketHandler
from litepipeline.manager.models.applications import Applications
from litepipeline.manager.models.tasks import Tasks
from litepipeline.manager.utils.scheduler import Scheduler
from litepipeline.manager.utils.common import file_sha1sum, file_md5sum, Errors, Stage, splitall, JSONLoadError
from litepipeline.manager.config import CONFIG

LOG = logging.getLogger("__name__")


class CreateTaskHandler(BaseHandler):
    @gen.coroutine
    def post(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            task_name = self.get_json_argument("task_name", "")
            app_id = self.get_json_argument("app_id", "")
            input_data = self.get_json_argument("input_data", {})
            if app_id and Applications.instance().get(app_id) and task_name:
                if not isinstance(input_data, dict):
                    raise JSONLoadError("input_data must be dict type")
                task_id = Tasks.instance().add(task_name, app_id, input_data = input_data)
                if task_id is not False:
                    result["task_id"] = task_id
                else:
                    Errors.set_result_error("OperationFailed", result)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
            LOG.debug("CreateTaskHandler, task_name: %s, app_id: %s", task_name, app_id)
        except JSONLoadError as e:
            LOG.error(e)
            Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class RunTaskHandler(BaseHandler):
    @gen.coroutine
    def put(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            task_id = self.get_json_argument("task_id", "")
            if task_id:
                task = Tasks.instance().get(task_id)
                if task:
                    if task["stage"] == Stage.finished:
                        success = Tasks.instance().update(task_id, {"stage": Stage.pending, "status": None})
                        if not success:
                            Errors.set_result_error("OperationFailed", result)
                    else:
                        Errors.set_result_error("TaskStillRunning", result)
                elif task is None:
                    Errors.set_result_error("TaskNotExists", result)
                else:
                    Errors.set_result_error("OperationFailed", result)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
            LOG.debug("RunTaskHandler, task_id: %s", task_id)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class StopTaskHandler(BaseHandler): # kill -9 or -15
    @gen.coroutine
    def put(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            task_id = self.get_json_argument("task_id", "")
            signal = int(self.get_json_argument("signal", -15))
            if task_id and signal in (-9, -15):
                task = Tasks.instance().get(task_id)
                if task:
                    if task["stage"] != Stage.finished:
                        success = yield Scheduler.instance().stop_task(task_id, signal)
                        if not success:
                            Errors.set_result_error("OperationFailed", result)
                    else:
                        Errors.set_result_error("TaskAlreadyFinished", result)
                elif task is None:
                    Errors.set_result_error("TaskNotExists", result)
                else:
                    Errors.set_result_error("OperationFailed", result)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
            LOG.debug("StopTaskHandler, task_id: %s, signal: %s", task_id, signal)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class DeleteTaskHandler(BaseHandler):
    @gen.coroutine
    def delete(self):
        result = {"result": Errors.OK}
        try:
            task_id = self.get_argument("task_id", "")
            if task_id:
                success = Tasks.instance().delete(task_id)
                if not success:
                    Errors.set_result_error("OperationFailed", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class DeleteTaskWorkspaceHandler(BaseHandler):
    @gen.coroutine
    def put(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            task_ids = self.get_json_argument("task_ids", [])
            if isinstance(task_ids, list) and task_ids:
                result = yield Scheduler.instance().delete_task_workspace(task_ids)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
            LOG.debug("DeleteTaskWorkspaceHandler, task_ids: %s", task_ids)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class PackTaskWorkspaceHandler(BaseHandler):
    @gen.coroutine
    def put(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            task_id = self.get_json_argument("task_id", "")
            name = self.get_json_argument("name", "")
            force = self.get_json_argument("force", False)
            if task_id and name:
                result = yield Scheduler.instance().pack_task_workspace(task_id, name, force)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
            LOG.debug("PackTaskWorkspaceHandler, task_id: %s, name: %s", task_id, name)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class DownloadTaskWorkspaceHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK}
        redirect_flag = False
        try:
            task_id = self.get_argument("task_id", "")
            name = self.get_argument("name", "")
            if task_id and name:
                r = yield Scheduler.instance().select_task_node_info(task_id, name)
                if "node_info" in r and r["node_info"] and r["task_info"]:
                    http_host = r["node_info"]["http_host"]
                    http_port = r["node_info"]["http_port"]
                    if http_host == "127.0.0.1":
                        host_parts = urllib.parse.urlsplit("//" + self.request.host)
                        http_host = host_parts.hostname
                    create_at = r["task_info"]["create_at"]
                    url = "http://%s:%s/workspace/download?task_id=%s&create_at=%s&name=%s" % (http_host, http_port, task_id, create_at, name)
                    self.redirect(url)
                    redirect_flag = True
                else:
                    result = r
            else:
                Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        if not redirect_flag:
            self.write(result)
            self.finish()


class InfoTaskHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK}
        try:
            task_id = self.get_argument("task_id", "")
            if task_id:
                task_info = Tasks.instance().get(task_id)
                if task_info:
                    result["task_info"] = task_info
                    if task_info["stage"] == Stage.running:
                        result["task_running_info"] = Scheduler.instance().get_running_actions(task_id)
                elif task_info is None:
                    Errors.set_result_error("TaskNotExists", result)
                else:
                    Errors.set_result_error("OperationFailed", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class ListTaskHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK}
        try:
            offset = int(self.get_argument("offset", "0"))
            limit = int(self.get_argument("limit", "0"))
            stage = self.get_argument("stage", "")
            LOG.debug("ListTaskHandler offset: %s, limit: %s, stage: %s", offset, limit, stage)
            result["tasks"] = Tasks.instance().list(offset = offset, limit = limit, stage = stage)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class CleanTaskHistoryHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        result = {}
        self.write(result)
        self.finish()


class UpdateActionHandler(BaseHandler):
    @gen.coroutine
    def put(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            name = self.get_json_argument("name", "")
            task_id = self.get_json_argument("task_id", "")
            stage = self.get_json_argument("stage", "")
            if task_id and name and stage:
                if stage == Stage.running:
                    Scheduler.instance().update_running_action(self.json_data)
                elif stage == Stage.finished:
                    Scheduler.instance().update_finish_action(self.json_data)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
            LOG.debug("UpdateActionHandler, action_name: %s, task_id: %s", name, task_id)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()



# -*- coding: utf-8 -*-

import os
import json
import logging
import datetime

from tornado.httpclient import AsyncHTTPClient, HTTPRequest
import tornado.ioloop
import tornado.web
from tornado import gen

from litepipeline.manager.models.applications import Applications
from litepipeline.manager.models.tasks import Tasks
from litepipeline.manager.models.schedules import Schedules
from litepipeline.manager.utils.listener import Connection
from litepipeline.manager.utils.common import Errors, Stage, Status, Event, OperationError
from litepipeline.manager.config import CONFIG
from litepipeline.manager import logger

LOG = logging.getLogger(__name__)


class Scheduler(object):
    _instance = None

    def __new__(cls, interval = 1):
        if not cls._instance:
            cls._instance = object.__new__(cls)
            cls._instance.interval = interval
            cls._instance.ioloop_service()
            cls._instance.running_actions = []
            cls._instance.pending_actions = []
            cls._instance.abandoned_actions = []
            cls._instance.tasks = {}
            cls._instance.async_client = AsyncHTTPClient()
            cls._instance.current_select_index = 0
        return cls._instance

    @classmethod
    def instance(cls):
        return cls._instance

    def ioloop_service(self):
        self.periodic_schedule = tornado.ioloop.PeriodicCallback(
            self.schedule_service, 
            self.interval * 1000
        )
        self.periodic_schedule.start()
        self.periodic_execute = tornado.ioloop.PeriodicCallback(
            self.execute_service, 
            self.interval * 1000
        )
        self.periodic_execute.start()
        self.periodic_crontab = tornado.ioloop.PeriodicCallback(
            self.crontab_service, 
            60 * 1000 # 1 min
        )
        self.periodic_crontab.start()

    @gen.coroutine
    def select_node(self):
        result = None
        try:
            LOG.debug("clients_dict: %s, total_action_slots: %s", Connection.clients_dict, Connection.total_action_slots)
            for node in Connection.clients:
                if "http_host" in node.info and "http_port" in node.info:
                    http_host = node.info["http_host"]
                    http_port = node.info["http_port"]
                    LOG.debug("checking node full, %s:%s", http_host, http_port)
                    url = "http://%s:%s/status/full" % (http_host, http_port)
                    request = HTTPRequest(url = url, method = "GET")
                    try:
                        r = yield self.async_client.fetch(request)
                        if r.code == 200:
                            data = json.loads(r.body.decode("utf-8"))
                            if data["result"] == Errors.OK:
                                if not data["full"]:
                                    result = node
                                    break
                        else:
                            LOG.error("checking node full failed, %s", url)
                    except ConnectionRefusedError as e:
                        LOG.warning("Scheduler.select_node: GET %s, %s", url, e)
        except Exception as e:
            LOG.exception(e)
        raise gen.Return(result)

    @gen.coroutine
    def select_node_balanced(self):
        result = None
        try:
            LOG.debug("clients_dict: %s, total_action_slots: %s", Connection.clients_dict, Connection.total_action_slots)
            node_ids_tmp = list(Connection.clients_dict.keys())
            node_ids_tmp.sort()
            node_ids = []
            if self.current_select_index >= len(node_ids_tmp):
                self.current_select_index = 0
            if self.current_select_index < 0:
                self.current_select_index = 0
            node_ids.extend(node_ids_tmp[self.current_select_index:])
            node_ids.extend(node_ids_tmp[:self.current_select_index])
            LOG.info("node_ids: %s, %s", node_ids, self.current_select_index)
            for i, node_id in enumerate(node_ids):
                node = Connection.clients_dict[node_id]
                if "http_host" in node.info and "http_port" in node.info:
                    http_host = node.info["http_host"]
                    http_port = node.info["http_port"]
                    LOG.debug("checking node full, %s:%s", http_host, http_port)
                    url = "http://%s:%s/status/full" % (http_host, http_port)
                    request = HTTPRequest(url = url, method = "GET")
                    try:
                        r = yield self.async_client.fetch(request)
                        if r.code == 200:
                            data = json.loads(r.body.decode("utf-8"))
                            if data["result"] == Errors.OK:
                                if not data["full"]:
                                    result = node
                                    self.current_select_index += 1
                                    break
                        else:
                            LOG.error("checking node full failed, %s", url)
                    except ConnectionRefusedError as e:
                        LOG.warning("Scheduler.select_node_balanced: GET %s, %s", url, e)
        except Exception as e:
            LOG.exception(e)
        raise gen.Return(result)

    @gen.coroutine
    def select_task_node_info(self, task_id, action_name):
        result = {"result": Errors.OK}
        try:
            task_info = Tasks.instance().get(task_id)
            if "result" in task_info and action_name in task_info["result"]:
                action_info = task_info["result"][action_name]
                if "node_id" in action_info and action_info["node_id"]:
                    if "stage" in action_info and action_info["stage"] == Stage.finished:
                        node_id = action_info["node_id"]
                        if node_id in Connection.clients_dict:
                            node = Connection.clients_dict[node_id]
                            result["node_info"] = node.info
                            result["task_info"] = task_info
                        else:
                            Errors.set_result_error("NodeNotExists", result)
                    else:
                        Errors.set_result_error("ActionNotFinished", result)
                else:
                    Errors.set_result_error("ActionNoNodeId", result)
            else:
                Errors.set_result_error("ActionNotRun", result)
        except Exception as e:
            LOG.exception(e)
        raise gen.Return(result)

    @gen.coroutine
    def delete_task_workspace(self, task_ids):
        result = {"result": Errors.OK, "failed_task_ids": []}
        try:
            task_infos = []
            for task_id in task_ids:
                task_info = Tasks.instance().get(task_id)
                task_infos.append({"task_id": task_id, "create_at": task_info["create_at"]})
            for node in Connection.clients:
                node_id = node.info["node_id"]
                http_host = node.info["http_host"]
                http_port = node.info["http_port"]
                LOG.debug("delete task_ids: %s workspace from node: %s(%s:%s)", task_ids, node_id, http_host, http_port)
                url = "http://%s:%s/workspace/delete" % (http_host, http_port)
                request = HTTPRequest(url = url, method = "PUT", body = json.dumps({"task_infos": task_infos}))
                r = yield self.async_client.fetch(request)
                if r.code == 200:
                    result_node = json.loads(r.body.decode("utf-8"))
                    if "failed_task_ids" in result_node:
                        result["failed_task_ids"].extend(result_node["failed_task_ids"])
                    LOG.debug("delete workspace from node: %s(%s:%s), result: %s", node_id, http_host, http_port, result_node)
                else:
                    Errors.set_result_error("OperationFailed", result)
                    LOG.warning("delete workspace from node: %s(%s:%s) failed", node_id, http_host, http_port)
                break
        except Exception as e:
            LOG.exception(e)
        raise gen.Return(result)

    @gen.coroutine
    def pack_task_workspace(self, task_id, action_name, force = False):
        result = {"result": Errors.OK}
        try:
            task_info = Tasks.instance().get(task_id)
            create_at = task_info["create_at"]
            if "result" in task_info and action_name in task_info["result"]:
                action_info = task_info["result"][action_name]
                if "node_id" in action_info and action_info["node_id"]:
                    if "stage" in action_info and action_info["stage"] == Stage.finished:
                        node_id = action_info["node_id"]
                        if node_id in Connection.clients_dict:
                            node = Connection.clients_dict[node_id]
                            http_host = node.info["http_host"]
                            http_port = node.info["http_port"]
                            LOG.debug("pack task_id: %s, action: %s workspace from node: %s(%s:%s)", task_id, action_name, node_id, http_host, http_port)
                            url = "http://%s:%s/workspace/pack" % (http_host, http_port)
                            request = HTTPRequest(url = url, method = "PUT", body = json.dumps({"task_id": task_id, "create_at": create_at, "name": action_name, "force": force}))
                            r = yield self.async_client.fetch(request)
                            if r.code == 200:
                                result = json.loads(r.body.decode("utf-8"))
                                LOG.debug("pack workspace from node: %s(%s:%s), result: %s", node_id, http_host, http_port, result)
                            else:
                                Errors.set_result_error("OperationFailed", result)
                                LOG.warning("pack workspace from node: %s(%s:%s) failed", node_id, http_host, http_port)
                        else:
                            Errors.set_result_error("NodeNotExists", result)
                    else:
                        Errors.set_result_error("ActionNotFinished", result)
                else:
                    Errors.set_result_error("ActionNoNodeId", result)
            else:
                Errors.set_result_error("ActionNotRun", result)
        except Exception as e:
            LOG.exception(e)
        raise gen.Return(result)

    def select_executable_action(self):
        result = None
        try:
            for action in self.pending_actions:
                if action["name"] in Event.events:
                    result = action
                    break
                elif set(action["condition"]).issubset(set(self.tasks[action["task_id"]]["finished"].keys())):
                    for condition_name in action["condition"]:
                        if "input_data" not in action:
                            action["input_data"] = {condition_name: self.tasks[action["task_id"]]["finished"][condition_name]["result"]["data"]}
                        else:
                            action["input_data"][condition_name] = self.tasks[action["task_id"]]["finished"][condition_name]["result"]["data"]
                    result = action
                    break
        except Exception as e:
            LOG.exception(e)
        return result

    def select_abandoned_action(self):
        result = None
        try:
            if self.abandoned_actions:
                result = self.abandoned_actions.pop(0)
        except Exception as e:
            LOG.exception(e)
        return result

    def can_load_more_task(self):
        result = False
        try:
            executable_action = self.select_executable_action()
            if executable_action is None and len(self.running_actions) < Connection.total_action_slots:
                result = True
        except Exception as e:
            LOG.exception(e)
        return result

    def schedule_finish_action(self):
        try:
            for action in self.running_actions:
                pass
        except Exception as e:
            LOG.exception(e)

    def get_running_actions(self, task_id):
        result = []
        try:
            for action in self.running_actions:
                if action["task_id"] == task_id:
                    result.append(action)
        except Exception as e:
            LOG.exception(e)
        return result

    def update_running_action(self, action_data):
        try:
            action_running = None
            now = datetime.datetime.now()
            task_id = action_data["task_id"]
            action_name = action_data["name"]
            for action in self.running_actions:
                if action["task_id"] == task_id and action["name"] == action_name:
                    action["update_at"] = str(now)
                    break
        except Exception as e:
            LOG.exception(e)

    def update_finish_action(self, action_result):
        try:
            action_finish = None
            task_id = action_result["task_id"]
            action_name = action_result["name"]
            now = datetime.datetime.now()
            for action in self.running_actions:
                if action["task_id"] == task_id and action["name"] == action_name:
                    action_finish = action
                    break
            if action_finish:
                if action_finish["name"] in Event.events: # event actions
                    task_info = Tasks.instance().get(task_id)
                    task_info["result"][action_finish["name"]] = action_result
                    Tasks.instance().update(task_id, {"result": task_info["result"]})
                    self.running_actions.remove(action_finish)
                else: # normal task actions
                    if action_result["status"] == Status.success:
                        if "actions" in action_result["result"]:
                            to_actions = {}
                            for action in action_result["result"]["actions"]:
                                self.tasks[task_id]["condition"].append(action["name"])
                                action["task_id"] = task_id
                                action["app_id"] = self.tasks[task_id]["app_info"]["application_id"]
                                action["app_sha1"] = self.tasks[task_id]["app_info"]["sha1"]
                                action["task_create_at"] = self.tasks[task_id]["task_info"]["create_at"]
                                if "to_action" in action:
                                    if action["to_action"] in to_actions:
                                        to_actions[action["to_action"]].append(action["name"])
                                    else:
                                        to_actions[action["to_action"]] = [action["name"]]
                                self.pending_actions.append(action)
                            for action in self.pending_actions:
                                if action["name"] in to_actions.keys():
                                    action["condition"].extend(to_actions[action["name"]])
                        self.tasks[task_id]["finished"][action_finish["name"]] = action_result
                        self.running_actions.remove(action_finish)
                        finish_condition = self.tasks[task_id]["condition"]
                        current_condition = self.tasks[task_id]["finished"].keys()
                        if len(current_condition) == len(finish_condition): # task finish & success
                            Tasks.instance().update(task_id, {"stage": Stage.finished, "status": Status.success, "end_at": now, "result": self.tasks[task_id]["finished"]})
                            if Event.success in self.tasks[task_id]["event_actions"]:
                                action = self.tasks[task_id]["event_actions"][Event.success]
                                action["input_data"] = {"result": self.tasks[task_id]["finished"]}
                                self.pending_actions.append(action)
                            del self.tasks[task_id]
                        else: # task running
                            Tasks.instance().update(task_id, {"result": self.tasks[task_id]["finished"]})
                    # action failed
                    else:
                        self.tasks[task_id]["finished"][action_finish["name"]] = action_result
                        pending_actions_tmp = []
                        running_actions_tmp = []
                        for action in self.pending_actions:
                            if action["task_id"] != task_id:
                                pending_actions_tmp.append(action)
                        self.pending_actions = pending_actions_tmp
                        for action in self.running_actions:
                            if action["task_id"] != task_id:
                                running_actions_tmp.append(action)
                            else:
                                self.abandoned_actions.append(action)
                        self.running_actions = running_actions_tmp
                        if "signal" in action_finish: # task failed by kill
                            Tasks.instance().update(task_id, {"stage": Stage.finished, "status": Status.kill, "end_at": now, "result": self.tasks[task_id]["finished"]})
                        else: # task failed
                            Tasks.instance().update(task_id, {"stage": Stage.finished, "status": Status.fail, "end_at": now, "result": self.tasks[task_id]["finished"]})
                        if Event.fail in self.tasks[task_id]["event_actions"]:
                            action = self.tasks[task_id]["event_actions"][Event.fail]
                            action["input_data"] = {"result": self.tasks[task_id]["finished"]}
                            self.pending_actions.append(action)
                        del self.tasks[task_id]
        except Exception as e:
            LOG.exception(e)

    @gen.coroutine
    def stop_task(self, task_id, signal):
        LOG.info("stop task_id: %s, signal: %s", task_id, signal)
        result = False
        try:
            now = datetime.datetime.now()
            pending_actions_tmp = []
            for action in self.pending_actions:
                if action["task_id"] != task_id:
                    pending_actions_tmp.append(action)
            self.pending_actions = pending_actions_tmp
            signal_actions = []
            for action in self.running_actions:
                if action["task_id"] == task_id:
                    action["signal"] = signal
                    signal_actions.append(action)
            if signal_actions:
                for action in signal_actions:
                    url = "http://%s/action/stop" % action["node"]
                    request = HTTPRequest(url = url, method = "PUT", body = json.dumps(action))
                    r = yield self.async_client.fetch(request)
                    if r.code == 200 and json.loads(r.body.decode("utf-8"))["result"] == Errors.OK:
                        LOG.debug("send signal[%s] to action[%s][%s] success", action["signal"], action["task_id"], action["name"])
                    else:
                        raise OperationError("request node[%s] for stopping action[%s][%s] failed" % (url, action["task_id"], action["name"]))
                self.tasks[task_id][Stage.stopping] = True
                Tasks.instance().update(task_id, {"stage": Stage.stopping})
            else:
                Tasks.instance().update(task_id, {"stage": Stage.finished, "status": Status.kill, "end_at": now, "result": self.tasks[task_id]["finished"]})
                del self.tasks[task_id]
            result = True
        except OperationError as e:
            LOG.error(e)
        except Exception as e:
            LOG.exception(e)
        return result

    @gen.coroutine
    def cancel_action(self, action):
        result = False
        try:
            url = "http://%s/action/cancel" % action["node"]
            request = HTTPRequest(url = url, method = "PUT", body = json.dumps(action))
            r = yield self.async_client.fetch(request)
            if r.code == 200 and json.loads(r.body.decode("utf-8"))["result"] == Errors.OK:
                LOG.debug("send cancel signal to action[%s][%s] success", action["task_id"], action["name"])
            else:
                raise OperationError("request node[%s] for cancelling action[%s][%s] failed" % (url, action["task_id"], action["name"]))
            result = True
        except OperationError as e:
            LOG.error(e)
        except Exception as e:
            LOG.exception(e)
        return result

    def repending_running_actions(self, node_id):
        LOG.warning("repending running actions on node[node_id: %s]", node_id)
        try:
            running_actions_tmp = []
            for action in self.running_actions:
                if action["node_id"] != node_id:
                    running_actions_tmp.append(action)
                else:
                    self.pending_actions.append(action)
            self.running_actions = running_actions_tmp
        except Exception as e:
            LOG.exception(e)

    @gen.coroutine
    def execute_service(self):
        LOG.debug("execute_service")
        try:
            node = yield self.select_node_balanced() 
            if node:
                http_host = node.info["http_host"]
                http_port = node.info["http_port"]
                LOG.debug("select node: %s:%s", http_host, http_port)
                action = self.select_executable_action()
                LOG.info("seleted action: %s", action)
                if action:
                    url = "http://%s:%s/action/run" % (http_host, http_port)
                    request = HTTPRequest(url = url, method = "POST", body = json.dumps(action))
                    r = yield self.async_client.fetch(request)
                    if r.code == 200 and json.loads(r.body.decode("utf-8"))["result"] == Errors.OK:
                        self.pending_actions.remove(action)
                        action["node"] = "%s:%s" % (http_host, http_port)
                        action["node_id"] = node.info["node_id"]
                        self.running_actions.append(action)
                        LOG.debug("migrate action[%s][%s] to running", action["task_id"], action["name"])
                    else:
                        LOG.error("request node[%s] for action[%s][%s] failed", url, action["task_id"], action["name"])
                else:
                    LOG.debug("no more executable action")
            else:
                LOG.warning("no selectable node")
            abandoned_action = self.select_abandoned_action()
            if abandoned_action:
                task_id = abandoned_action["task_id"]
                app_id = abandoned_action["app_id"]
                action_name = abandoned_action["name"]
                success = yield self.cancel_action(abandoned_action)
                if not success:
                    self.abandoned_actions.append(abandoned_action)
                LOG.info("abandon action, app_id: %s, task_id: %s, action: %s, success: %s", app_id, task_id, action_name, success)
        except Exception as e:
            LOG.exception(e)

    @gen.coroutine
    def schedule_service(self):
        LOG.debug("schedule_service")
        try:
            load_more_task = self.can_load_more_task()
            LOG.debug("can load more task: %s", load_more_task)
            if load_more_task:
                task_info = Tasks.instance().get_first()
                if task_info:
                    task_id = task_info["task_id"]
                    app_id = task_info["application_id"]
                    app_info = Applications.instance().get(app_id)
                    if app_info:
                        app_config_path = os.path.join(CONFIG["data_path"], "applications", app_id[:2], app_id[2:4], app_id, "app", "configuration.json")
                        if os.path.exists(app_config_path):
                            fp = open(app_config_path, "r")
                            app_config = json.loads(fp.read())
                            fp.close()
                            finish_condition = []
                            for action in app_config["actions"]:
                                action["task_id"] = task_id
                                action["app_id"] = app_id
                                action["app_sha1"] = app_info["sha1"]
                                action["task_create_at"] = task_info["create_at"]
                                if len(action["condition"]) == 0:
                                    action["input_data"] = task_info["input_data"]
                                finish_condition.append(action["name"])
                                self.pending_actions.append(action)
                            event_actions = {}
                            if "event_actions" in app_config:
                                for event_name in app_config["event_actions"]:
                                    action = app_config["event_actions"][event_name]
                                    action["task_id"] = task_id
                                    action["app_id"] = app_id
                                    action["app_sha1"] = app_info["sha1"]
                                    action["task_create_at"] = task_info["create_at"]
                                event_actions = app_config["event_actions"]
                            self.tasks[task_id] = {"task_info": task_info, "condition": finish_condition, "app_info": app_info, "finished": {}, "event_actions": event_actions}
                            Tasks.instance().update(task_id, {"stage": Stage.running, "start_at": datetime.datetime.now()})
                        else:
                            LOG.error("Scheduler app config file[%s] not exists", app_config_path)
                    elif app_info is None:
                        LOG.error("Scheduler task[%s]'s app_info[%s] not exists", task_id, app_id)
                    else:
                        LOG.error("Scheduler get task[%s]'s app_info[%s] failed", task_id, app_id)
                elif task_info is None:
                    LOG.debug("Scheduler no more task to execute")
                else:
                    LOG.error("Scheduler get task failed")
            else:
                LOG.warning("no selectable node")
        except Exception as e:
            LOG.exception(e)

    @gen.coroutine
    def crontab_service(self):
        LOG.debug("crontab_service")
        try:
            schedules = Schedules.instance()
            now = datetime.datetime.now()
            for schedule_id in schedules.cache:
                schedule = schedules.cache[schedule_id]
                executed = False if ("executed" not in schedule) or ("executed" in schedule and schedule["executed"] != str(now.date())) else True
                if not executed:
                    if schedule["day_of_month"] == -1 and schedule["day_of_week"] == -1:
                        hour = 0 if schedule["hour"] == -1 else schedule["hour"]
                        minute = 0 if schedule["minute"] == -1 else schedule["minute"]
                        if now.hour == hour and now.minute >= minute and now.minute - minute < 5:
                            task_id = Tasks.instance().add(schedule["schedule_name"], schedule["application_id"], stage = Stage.pending, input_data = schedule["input_data"])
                            if task_id:
                                schedule["executed"] = str(now.date())
                            else:
                                LOG.error("execute schedule every day failed: %s, at: %s, task_id: %s", schedule, now, task_id)
                            LOG.debug("execute schedule every day: %s, at: %s, task_id: %s", schedule, now, task_id)
                    elif schedule["day_of_month"] != -1 and schedule["day_of_month"] == now.day:
                        hour = 0 if schedule["hour"] == -1 else schedule["hour"]
                        minute = 0 if schedule["minute"] == -1 else schedule["minute"]
                        if now.hour == hour and now.minute >= minute and now.minute - minute < 5:
                            task_id = Tasks.instance().add(schedule["schedule_name"], schedule["application_id"], stage = Stage.pending, input_data = schedule["input_data"])
                            if task_id:
                                schedule["executed"] = str(now.date())
                            else:
                                LOG.error("execute schedule every month failed: %s, at: %s, task_id: %s", schedule, now, task_id)
                            LOG.debug("execute schedule every month: %s, at: %s, task_id: %s", schedule, now, task_id)
                    elif schedule["day_of_week"] != -1 and schedule["day_of_week"] == now.isoweekday():
                        hour = 0 if schedule["hour"] == -1 else schedule["hour"]
                        minute = 0 if schedule["minute"] == -1 else schedule["minute"]
                        if now.hour == hour and now.minute >= minute and now.minute - minute < 5:
                            task_id = Tasks.instance().add(schedule["schedule_name"], schedule["application_id"], stage = Stage.pending, input_data = schedule["input_data"])
                            if task_id:
                                schedule["executed"] = str(now.date())
                            else:
                                LOG.error("execute schedule every week failed: %s, at: %s, task_id: %s", schedule, now, task_id)
                            LOG.debug("execute schedule every week: %s, at: %s, task_id: %s", schedule, now, task_id)
                else:
                    LOG.debug("executed schedule: %s", schedule)
        except Exception as e:
            LOG.exception(e)

    def close(self):
        try:
            if self.periodic_schedule:
                self.periodic_schedule.stop()
            if self.periodic_execute:
                self.periodic_execute.stop()
            if self.periodic_crontab:
                self.periodic_crontab.stop()
            for task_id in self.tasks:
                Tasks.instance().update(task_id, {"stage": Stage.pending})
            LOG.debug("Scheduler close")
        except Exception as e:
            LOG.exception(e)



# -*- coding: utf-8 -*-

import os
import time
import json
import hashlib
import logging

from tornado import ioloop
from tornado import gen

from litepipeline.manager.config import CONFIG

LOG = logging.getLogger(__name__)

BUF_SIZE = 65536


class Servers(object):
    HTTP_SERVER = None
    DB_SERVERS = []
    TORNADO_INSTANCE = None


async def shutdown():
    LOG.info("Stopping Service(%s:%s)", CONFIG["http_host"], CONFIG["http_port"])
    if Servers.HTTP_SERVER:
        Servers.HTTP_SERVER.stop()
        LOG.info("Stop http server!")
    for db_server in Servers.DB_SERVERS:
        db_server.close()
        LOG.info("Stop db server!")
    await gen.sleep(1)
    LOG.info("Will shutdown ...")
    ioloop.IOLoop.current().stop()


def sig_handler(sig, frame):
    LOG.warning("sig_handler Caught signal: %s", sig)
    ioloop.IOLoop.current().add_callback_from_signal(shutdown)


class Errors(object):
    OK = "ok"
    errors = {
        "ServerException": {"name": "ServerException", "message": "server exception"},
        "InvalidParameters": {"name": "InvalidParameters", "message": "invalid parameters"},
        "OperationFailed": {"name": "OperationFailed", "message": "operation failed"},
        "AppNotExists": {"name": "AppNotExists", "message": "application not exists"},
        "TaskNotExists": {"name": "TaskNotExists", "message": "task not exists"},
        "TaskAlreadyFinished": {"name": "TaskAlreadyFinished", "message": "task already finished"},
        "TaskStillRunning": {"name": "TaskStillRunning", "message": "task still running"},
        "NodeNotExists": {"name": "NodeNotExists", "message": "node not exists"},
        "ActionNotFinished": {"name": "ActionNotFinished", "message": "action not finished"},
        "ActionNoNodeId": {"name": "ActionNoNodeId", "message": "action no node id"},
        "ActionNotRun": {"name": "ActionNotRun", "message": "action not run"},
        "ScheduleNotExists": {"name": "ScheduleNotExists", "message": "schedule no exists"},
    }

    @classmethod
    def set_result_error(cls, error_name, result, message = ""):
        if error_name in cls.errors:
            result["result"] = error_name
            if message == "":
                result["message"] = cls.errors[error_name]["message"]
            else:
                result["message"] = message
        else:
            result["result"] = "UnknownError"
            result["message"] = "unknown error"


class Stage(object):
    pending = "pending"
    running = "running"
    finished = "finished"
    stopping = "stopping"


class Status(object):
    fail = "fail"
    success = "success"
    kill = "kill"
    cancel = "cancel"
    terminate = "terminate"


class Event(object):
    events = ["fail", "success"]
    fail = "fail"
    success = "success"


class JSONLoadError(Exception):
    def __init__(self, message):
        self.message = message


class MetaNotDictError(Exception):
    def __init__(self, message):
        self.message = message


class OperationError(Exception):
    def __init__(self, message):
        self.message = message


def file_sha1sum(file_path):
    sha1 = hashlib.sha1()
    with open(file_path, 'rb') as f:
        while True:
            data = f.read(BUF_SIZE)
            if not data:
                break
            sha1.update(data)
    return sha1.hexdigest()


def file_md5sum(file_path):
    md5 = hashlib.md5()
    with open(file_path, 'rb') as fp:
        while True:
            data = fp.read(BUF_SIZE)
            if not data:
                break
            md5.update(data)
    return md5.hexdigest()


def splitall(path):
    allparts = []
    while True:
        parts = os.path.split(path)
        if parts[0] == path:  # sentinel for absolute paths
            allparts.insert(0, parts[0])
            break
        elif parts[1] == path: # sentinel for relative paths
            allparts.insert(0, parts[1])
            break
        else:
            path = parts[0]
            allparts.insert(0, parts[1])
    return allparts


def init_storage():
    directories = [
        os.path.join(CONFIG["data_path"], "applications"),
        os.path.join(CONFIG["data_path"], "tmp"),
    ]
    for d in directories:
        if not os.path.exists(d) or not os.path.isdir(d):
            os.makedirs(d)



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-

import os
import re
import sys
import json
import time
import argparse

import requests
from progress.spinner import Spinner

from litepipeline.version import __version__

parser = argparse.ArgumentParser(prog = 'litepipeline')

# common arguments
parser.add_argument("address", help = "manager address, host:port")
parser.add_argument("-w", "--column_width", help = "column max width", type = int, default = 0)
parser.add_argument("-v", "--version", action = 'version', version = '%(prog)s ' + __version__)
subparsers = parser.add_subparsers(dest = "object", help = 'sub-command help')

# operate with application
parser_app = subparsers.add_parser("app", help = "operate with app API")
subparsers_app = parser_app.add_subparsers(dest = "operation", help = 'sub-command app help')

parser_app_create = subparsers_app.add_parser("create", help = "create application")
parser_app_create.add_argument("-f", "--file", required = True, help = "application's file", default = "")
parser_app_create.add_argument("-n", "--name", required = True, help = "application's name", default = "")
parser_app_create.add_argument("-d", "--description", help = "application's description", default = "")
parser_app_create.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_app_delete = subparsers_app.add_parser("delete", help = "delete application")
parser_app_delete.add_argument("-a", "--app_id", required = True, help = "application id", default = "")
parser_app_delete.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_app_update = subparsers_app.add_parser("update", help = "update application")
parser_app_update.add_argument("-a", "--app_id", required = True, help = "application id", default = "")
parser_app_update.add_argument("-f", "--file", help = "application's file", default = "")
parser_app_update.add_argument("-n", "--name", help = "application's name", default = "")
parser_app_update.add_argument("-d", "--description", help = "application's description", default = "")
parser_app_update.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_app_list = subparsers_app.add_parser("list", help = "list applications")
parser_app_list.add_argument("-o", "--offset", help = "list offset", type = int, default = 0)
parser_app_list.add_argument("-l", "--limit", help = "list limit", type = int, default = 0)
parser_app_list.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_app_info = subparsers_app.add_parser("info", help = "application's info")
parser_app_info.add_argument("-a", "--app_id", required = True, help = "application id", default = "")
parser_app_info.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_app_download = subparsers_app.add_parser("download", help = "download application")
parser_app_download.add_argument("-a", "--app_id", required = True, help = "application id", default = "")
parser_app_download.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

# operate with task
parser_task = subparsers.add_parser("task", help = "operate with task API")
subparsers_task = parser_task.add_subparsers(dest = "operation", help = 'sub-command task help')

parser_task_create = subparsers_task.add_parser("create", help = "create task")
parser_task_create.add_argument("-a", "--app_id", required = True, help = "application id", default = "")
parser_task_create.add_argument("-n", "--name", required = True, help = "task's name", default = "")
parser_task_create.add_argument("-i", "--input", help = "task's input data, json string", default = "{}")
parser_task_create.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_task_delete = subparsers_task.add_parser("delete", help = "delete task")
parser_task_delete.add_argument("-t", "--task_id", required = True, help = "task id", default = "")
parser_task_delete.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_task_list = subparsers_task.add_parser("list", help = "list tasks")
parser_task_list.add_argument("-o", "--offset", help = "list offset", type = int, default = 0)
parser_task_list.add_argument("-l", "--limit", help = "list limit", type = int, default = 0)
parser_task_list.add_argument("-s", "--stage", choices = ["pending", "running", "finished"], help = "task's executing stage", default = "")
parser_task_list.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_task_info = subparsers_task.add_parser("info", help = "task's info")
parser_task_info.add_argument("-t", "--task_id", required = True, help = "task id", default = "")
parser_task_info.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_task_rerun = subparsers_task.add_parser("rerun", help = "rerun task")
parser_task_rerun.add_argument("-t", "--task_id", required = True, help = "task id", default = "")
parser_task_rerun.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_task_stop = subparsers_task.add_parser("stop", help = "stop task")
parser_task_stop.add_argument("-t", "--task_id", required = True, help = "task id", default = "")
parser_task_stop.add_argument("-g", "--signal", help = "stop task's signal: -9 or -15", type = int, default = -15)
parser_task_stop.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

# operate with cluster
parser_cluster = subparsers.add_parser("cluster", help = "operate with cluster API")
subparsers_cluster = parser_cluster.add_subparsers(dest = "operation", help = 'sub-command cluster help')

parser_cluster_info = subparsers_cluster.add_parser("info", help = "cluster's info")
parser_cluster_info.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

# operate with workspace
parser_workspace = subparsers.add_parser("workspace", help = "operate with workspace API")
subparsers_workspace = parser_workspace.add_subparsers(dest = "operation", help = 'sub-command workspace help')

parser_workspace_delete = subparsers_workspace.add_parser("delete", help = "delete workspace")
parser_workspace_delete.add_argument("-t", "--task_id", required = True, help = "task id", action = "append")
parser_workspace_delete.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_workspace_download = subparsers_workspace.add_parser("download", help = "download workspace")
parser_workspace_download.add_argument("-t", "--task_id", required = True, help = "task id", default = "")
parser_workspace_download.add_argument("-n", "--name", required = True, help = "action name", default = "")
parser_workspace_download.add_argument("-f", "--force", help = "force repack workspace", action = "store_true")
parser_workspace_download.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

# operate with schedule
parser_schedule = subparsers.add_parser("schedule", help = "operate with schedule API")
subparsers_schedule = parser_schedule.add_subparsers(dest = "operation", help = 'sub-command schedule help')

parser_schedule_create = subparsers_schedule.add_parser("create", help = "create schedule")
parser_schedule_create.add_argument("-a", "--app_id", required = True, help = "application id", default = "")
parser_schedule_create.add_argument("-n", "--name", required = True, help = "schedule's name", default = "")
parser_schedule_create.add_argument("-i", "--input", help = "task's input data, json string", default = "{}")
parser_schedule_create.add_argument("-m", "--minute", help = "minute, [0, 59]", type = int, default = -1)
parser_schedule_create.add_argument("-H", "--hour", help = "hour, [0, 23]", type = int, default = -1)
parser_schedule_create.add_argument("-d", "--day_of_month", help = "day of month, [1, 31]", type = int, default = -1)
parser_schedule_create.add_argument("-D", "--day_of_week", help = "day of week, [0, 6] (Sunday = 0)", type = int, default = -1)
parser_schedule_create.add_argument("-e", "--enable", choices = ["true", "false"], help = "schedule's enable flag", default = "false")
parser_schedule_create.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_schedule_update = subparsers_schedule.add_parser("update", help = "update schedule")
parser_schedule_update.add_argument("-s", "--schedule_id", required = True, help = "schedule id", default = "")
parser_schedule_update.add_argument("-a", "--app_id", help = "application id")
parser_schedule_update.add_argument("-n", "--name", help = "schedule's name")
parser_schedule_update.add_argument("-i", "--input", help = "task's input data, json string")
parser_schedule_update.add_argument("-m", "--minute", help = "minute, [0, 59]", type = int)
parser_schedule_update.add_argument("-H", "--hour", help = "hour, [0, 23]", type = int)
parser_schedule_update.add_argument("-d", "--day_of_month", help = "day of month, [1, 31]", type = int)
parser_schedule_update.add_argument("-D", "--day_of_week", help = "day of week, [0, 6] (Sunday = 0)", type = int)
parser_schedule_update.add_argument("-e", "--enable", choices = ["true", "false"], help = "schedule's enable flag")
parser_schedule_update.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_schedule_delete = subparsers_schedule.add_parser("delete", help = "delete schedule")
parser_schedule_delete.add_argument("-s", "--schedule_id", required = True, help = "schedule id", default = "")
parser_schedule_delete.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_schedule_list = subparsers_schedule.add_parser("list", help = "list schedules")
parser_schedule_list.add_argument("-o", "--offset", help = "list offset", type = int, default = 0)
parser_schedule_list.add_argument("-l", "--limit", help = "list limit", type = int, default = 0)
parser_schedule_list.add_argument("-e", "--enable", choices = ["true", "false"], help = "schedule's enable flag", default = "")
parser_schedule_list.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

parser_schedule_info = subparsers_schedule.add_parser("info", help = "schedule's info")
parser_schedule_info.add_argument("-s", "--schedule_id", required = True, help = "schedule id", default = "")
parser_schedule_info.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

args = parser.parse_args()


def print_table_result(data, fields):
    fields.insert(0, "#")
    field_length_map = {}
    lines = []
    num = 1
    column_max_width = args.column_width
    for field in fields:
        field_length_map[field] = len(field)
    for item in data:
        line = []
        for field in fields:
            if field == "#":
                line.append(str(num))
            else:
                v = str(item[field]) if field in item else ""
                v_len = len(v)
                if column_max_width > 0 and v_len > column_max_width:
                    v_len = column_max_width
                    v = v[:v_len]
                if v_len > field_length_map[field]:
                    field_length_map[field] = v_len
                line.append(v)
        lines.append(tuple(line))
        num += 1
    field_length_map["#"] = len(str(num))
    format_str = ""
    for field in fields:
        field_len = field_length_map[field]
        if field == "#":
            format_str += "%" + " %s" % field_len + "s | "
        else:
            format_str += "%" + "-%s" % field_len + "s | "
    format_str = format_str[:-3]
    print(format_str % tuple(fields))
    for line in lines:
        print(format_str % line)


def main():
    try:
        address = args.address
        object = args.object
        operation = args.operation
        raw = args.raw
        url = "http://%s/%s/%s" % (address, object, operation)
        if address:
            if object == "app":
                if operation == "list":
                    url += "?offset=%s&limit=%s" % (args.offset, args.limit)
                    r = requests.get(url)
                    if r.status_code == 200:
                        data = r.json()
                        if raw:
                            print(json.dumps(data, indent = 4, sort_keys = True))
                        else:
                            if data["result"] == "ok": 
                                print_table_result(
                                    data["apps"],
                                    [
                                        "application_id",
                                        "name",
                                        "create_at",
                                        "update_at",
                                    ]
                                )
                            else:
                                print_table_result(
                                    [data],
                                    ["result", "message"]
                                )
                    else:
                        print("error:\ncode: %s\ncontent: %s" % (r.status_code, r.content))
                elif operation == "info":
                    if args.app_id:
                        url += "?app_id=%s" % args.app_id
                        r = requests.get(url)
                        if r.status_code == 200:
                            data = r.json()
                            if raw:
                                print(json.dumps(data, indent = 4, sort_keys = True))
                            else:
                                if data["result"] == "ok":
                                    print_table_result(
                                        [data["app_info"]],
                                        [
                                            "application_id",
                                            "name",
                                            "create_at",
                                            "update_at",
                                            "sha1",
                                            "description",
                                        ]
                                    )
                                else:
                                    print_table_result(
                                        [data],
                                        ["result", "message"]
                                    )
                        else:
                            print("error:\ncode: %s\ncontent: %s" % (r.status_code, r.content))
                    else:
                        parser.print_help()
                elif operation == "delete":
                    if args.app_id:
                        url += "?app_id=%s" % args.app_id
                        r = requests.delete(url)
                        if r.status_code == 200:
                            data = r.json()
                            if raw:
                                print(json.dumps(data, indent = 4, sort_keys = True))
                            else:
                                print_table_result(
                                    [data],
                                    ["result", "message"]
                                )
                        else:
                            print("error:\ncode: %s\ncontent: %s" % (r.status_code, r.content))
                    else:
                        parser.print_help()
                elif operation == "create":
                    if args.file and args.name:
                        if os.path.exists(args.file) and os.path.isfile(args.file):
                            file_name = os.path.split(args.file)[-1]
                            files = {'up_file': (file_name, open(args.file, "rb"), b"text/plain")}
                            values = {"name": args.name, "description": args.description}
                            r = requests.post(url, files = files, data = values)
                            if r.status_code == 200:
                                data = r.json()
                                if raw:
                                    print(json.dumps(data, indent = 4, sort_keys = True))
                                else:
                                    print_table_result(
                                        [data],
                                        ["app_id", "result", "message"]
                                    )
                            else:
                                print("error:\ncode: %s\ncontent: %s" % (r.status_code, r.content))
                        else:
                            print("file[%s] not exists")
                    else:
                        parser.print_help()
                elif operation == "update":
                    if args.app_id:
                        files = {}
                        values = {"app_id": args.app_id}
                        executable = True
                        if args.file:
                            if os.path.exists(args.file) and os.path.isfile(args.file):
                                file_name = os.path.split(args.file)[-1]
                                files = {'up_file': (file_name, open(args.file, "rb"), b"text/plain")}
                            else:
                                print("file[%s] not exists")
                                executable = False
                        if args.name:
                            values["name"] = args.name
                        if args.description:
                            values["description"] = args.description
                        if executable:
                            if files:
                                r = requests.put(url, files = files, data = values)
                            else:
                                for key in values:
                                    values[key] = (None, values[key])
                                r = requests.put(url, files = values)
                            if r.status_code == 200:
                                data = r.json()
                                if raw:
                                    print(json.dumps(data, indent = 4, sort_keys = True))
                                else:
                                    print_table_result(
                                        [data],
                                        ["app_id", "result", "message"]
                                    )
                            else:
                                print("error:\ncode: %s\ncontent: %s" % (r.status_code, r.content))
                    else:
                        print("error: need app_id(-a, --app_id) parameter")
                elif operation == "download":
                    if args.app_id:
                        url += "?app_id=%s" % args.app_id
                        file_path = "./%s.tar.gz" % args.app_id
                        r = requests.get(url)
                        if r.status_code == 200:
                            f = open(file_path, 'wb')
                            f.write(r.content)
                            f.close()
                            print("application: %s" % file_path)
                        else:
                            print("error:\ncode: %s\ncontent: %s" % (r.status_code, r.content))
                    else:
                        print("error: need app_id(-a, --app_id) parameter")    
            elif object == "task":
                if operation == "list":
                    url += "?offset=%s&limit=%s" % (args.offset, args.limit)
                    if args.stage:
                        url += "&stage=%s" % args.stage
                    r = requests.get(url)
                    if r.status_code == 200:
                        data = r.json()
                        if raw:
                            print(json.dumps(data, indent = 4, sort_keys = True))
                        else:
                            if data["result"] == "ok":
                                print_table_result(
                                    data["tasks"],
                                    [
                                        "task_id",
                                        "application_id",
                                        "task_name",
                                        "create_at",
                                        "start_at",
                                        "end_at",
                                        "stage",
                                        "status",
                                    ]
                                )
                            else:
                                print_table_result(
                                    [data],
                                    ["result", "message"]
                                )
                    else:
                        print("error:\ncode: %s\ncontent: %s" % (r.status_code, r.content))
                elif operation == "info":
                    if args.task_id:
                        url += "?task_id=%s" % args.task_id
                        r = requests.get(url)
                        if r.status_code == 200:
                            data = r.json()
                            if raw:
                                print(json.dumps(data, indent = 4, sort_keys = True))
                            else:
                                if data["result"] == "ok":
                                    print_table_result(
                                        [data["task_info"]],
                                        [
                                            "task_id",
                                            "application_id",
                                            "task_name",
                                            "create_at",
                                            "start_at",
                                            "end_at",
                                            "stage",
                                            "status",
                                        ]
                                    )
                                else:
                                    print_table_result(
                                        [data],
                                        ["result", "message"]
                                    )
                        else:
                            print("error:\ncode: %s\ncontent: %s" % (r.status_code, r.content))
                    else:
                        parser.print_help()
                elif operation == "delete":
                    if args.task_id:
                        url += "?task_id=%s" % args.task_id
                        r = requests.delete(url)
                        if r.status_code == 200:
                            data = r.json()
                            if raw:
                                print(json.dumps(data, indent = 4, sort_keys = True))
                            else:
                                print_table_result(
                                    [data],
                                    ["result", "message"]
                                )
                        else:
                            print("error:\ncode: %s\ncontent: %s" % (r.status_code, r.content))
                    else:
                        parser.print_help()
                elif operation == "create":
                    if args.name and args.app_id and args.input:
                        try:
                            data = {"task_name": args.name, "app_id": args.app_id, "input_data": json.loads(args.input)}
                            r = requests.post(url, json = data)
                            if r.status_code == 200:
                                data = r.json()
                                if raw:
                                    print(json.dumps(data, indent = 4, sort_keys = True))
                                else:
                                    print_table_result(
                                        [data],
                                        ["task_id", "result", "message"]
                                    )
                            else:
                                print("error:\ncode: %s\ncontent: %s" % (r.status_code, r.content))
                        except Exception as e:
                            print(e)
                    else:
                        parser.print_help()
                elif operation == "rerun":
                    if args.task_id:
                        try:
                            data = {"task_id": args.task_id}
                            r = requests.put(url, json = data)
                            if r.status_code == 200:
                                data = r.json()
                                if raw:
                                    print(json.dumps(data, indent = 4, sort_keys = True))
                                else:
                                    print_table_result(
                                        [data],
                                        ["result", "message"]
                                    )
                            else:
                                print("error:\ncode: %s\ncontent: %s" % (r.status_code, r.content))
                        except Exception as e:
                            print(e)
                    else:
                        parser.print_help()
                elif operation == "stop":
                    if args.task_id and args.signal:
                        try:
                            data = {"task_id": args.task_id, "signal": args.signal}
                            r = requests.put(url, json = data)
                            if r.status_code == 200:
                                data = r.json()
                                if raw:
                                    print(json.dumps(data, indent = 4, sort_keys = True))
                                else:
                                    print_table_result(
                                        [data],
                                        ["result", "message"]
                                    )
                            else:
                                print("error:\ncode: %s\ncontent: %s" % (r.status_code, r.content))
                        except Exception as e:
                            print(e)
                    else:
                        parser.print_help()
            elif object == "cluster":
                if operation == "info":
                    r = requests.get(url)
                    if r.status_code == 200:
                        data = r.json()
                        if raw:
                            print(json.dumps(data, indent = 4, sort_keys = True))
                        else:
                            if data["result"] == "ok":
                                print_table_result(
                                    data["info"]["nodes"],
                                    [
                                        "node_id",
                                        "http_host",
                                        "http_port",
                                        "action_slots",
                                        "app_path",
                                        "data_path",
                                    ]
                                )
                            else:
                                print_table_result(
                                    [data],
                                    ["result", "message"]
                                )
                    else:
                        print("error:\ncode: %s\ncontent: %s" % (r.status_code, r.content))
            elif object == "workspace":
                if operation == "delete":
                    if args.task_id:
                        try:
                            data = {"task_ids": args.task_id}
                            r = requests.put(url, json = data)
                            if r.status_code == 200:
                                data = r.json()
                                if raw:
                                    print(json.dumps(data, indent = 4, sort_keys = True))
                                else:
                                    print_table_result(
                                        [data],
                                        ["result", "message"]
                                    )
                            else:
                                print("error:\ncode: %s\ncontent: %s" % (r.status_code, r.content))
                        except Exception as e:
                            print(e)
                    else:
                        parser.print_help()
                elif operation == "download":
                    if args.task_id and args.name:
                        pack_error = False
                        download_ready = False
                        try:
                            url = "http://%s/%s/%s" % (address, object, "pack")
                            spinner = Spinner('Packing ... ')
                            if args.force:
                                data = {"task_id": args.task_id, "name": args.name, "force": args.force}
                                r = requests.put(url, json = data)
                                if r.status_code == 200:
                                    data = r.json()
                                    if "result" in data and data["result"] == "ok":
                                        download_ready = True
                                    elif "result" in data and data["result"] == "OperationRunning":
                                        spinner.next()
                                    else:
                                        pack_error = True
                                        if raw:
                                            print(json.dumps(data, indent = 4, sort_keys = True))
                                        else:
                                            print_table_result(
                                                [data],
                                                ["result", "message"]
                                            )
                                else:
                                    pack_error = True
                                    print("error:\ncode: %s\ncontent: %s" % (r.status_code, r.content))
                                time.sleep(0.5)
                            while not pack_error and not download_ready:
                                data = {"task_id": args.task_id, "name": args.name, "force": False}
                                r = requests.put(url, json = data)
                                if r.status_code == 200:
                                    data = r.json()
                                    if "result" in data and data["result"] == "ok":
                                        download_ready = True
                                    elif "result" in data and data["result"] == "OperationRunning":
                                        spinner.next()
                                    else:
                                        pack_error = True
                                        if raw:
                                            print(json.dumps(data, indent = 4, sort_keys = True))
                                        else:
                                            print_table_result(
                                                [data],
                                                ["result", "message"]
                                            )
                                else:
                                    pack_error = True
                                    print("error:\ncode: %s\ncontent: %s" % (r.status_code, r.content))
                                time.sleep(0.5)
                            url = "http://%s/%s/%s" % (address, object, "download")
                            url += "?task_id=%s&name=%s" % (args.task_id, args.name)
                            print("\nDownload from: %s" % url)
                            spinner = Spinner('Downloading ... ')
                            if not pack_error and download_ready:
                                with requests.get(url, allow_redirects = True, stream = True, timeout = 3600) as r:
                                    r.raise_for_status()
                                    file_name = "%s.%s.tar.gz" % (args.task_id, args.name)
                                    file_path = os.path.join("./", file_name)
                                    with open(file_path, "wb") as f:
                                        for chunk in r.iter_content(chunk_size = 64 * 1024):
                                            if chunk:
                                                f.write(chunk)
                                                spinner.next()
                                    print("\nWorkspace: %s" % file_path)
                        except Exception as e:
                            print(e)
                    else:
                        parser.print_help()
            elif object == "schedule":
                if operation == "list":
                    url += "?offset=%s&limit=%s" % (args.offset, args.limit)
                    if args.enable:
                        url += "&enable=%s" % args.enable
                    r = requests.get(url)
                    if r.status_code == 200:
                        data = r.json()
                        if raw:
                            print(json.dumps(data, indent = 4, sort_keys = True))
                        else:
                            if data["result"] == "ok":
                                print_table_result(
                                    data["schedules"],
                                    [
                                        "schedule_id",
                                        "application_id",
                                        "schedule_name",
                                        "create_at",
                                        "update_at",          
                                        "hour",
                                        "minute",
                                        "day_of_month",
                                        "day_of_week",
                                        "enable",
                                    ]
                                )
                            else:
                                print_table_result(
                                    [data],
                                    ["result", "message"]
                                )
                    else:
                        print("error:\ncode: %s\ncontent: %s" % (r.status_code, r.content))
                elif operation == "info":
                    if args.schedule_id:
                        url += "?schedule_id=%s" % args.schedule_id
                        r = requests.get(url)
                        if r.status_code == 200:
                            data = r.json()
                            if raw:
                                print(json.dumps(data, indent = 4, sort_keys = True))
                            else:
                                if data["result"] == "ok":
                                    print_table_result(
                                        [data["schedule_info"]],
                                        [
                                            "schedule_id",
                                            "application_id",
                                            "schedule_name",
                                            "create_at",
                                            "update_at",
                                            "hour",
                                            "minute",
                                            "day_of_month",
                                            "day_of_week",
                                            "enable",
                                        ]
                                    )
                                else:
                                    print_table_result(
                                        [data],
                                        ["result", "message"]
                                    )
                        else:
                            print("error:\ncode: %s\ncontent: %s" % (r.status_code, r.content))
                    else:
                        parser.print_help()
                elif operation == "delete":
                    if args.schedule_id:
                        url += "?schedule_id=%s" % args.schedule_id
                        r = requests.delete(url)
                        if r.status_code == 200:
                            data = r.json()
                            if raw:
                                print(json.dumps(data, indent = 4, sort_keys = True))
                            else:
                                print_table_result(
                                    [data],
                                    ["result", "message"]
                                )
                        else:
                            print("error:\ncode: %s\ncontent: %s" % (r.status_code, r.content))
                    else:
                        parser.print_help()
                elif operation == "create":
                    if args.name and args.app_id and args.input:
                        try:
                            data = {
                                "schedule_name": args.name,
                                "app_id": args.app_id,
                                "input_data": json.loads(args.input),
                                "minute": args.minute,
                                "hour": args.hour,
                                "day_of_month": args.day_of_month,
                                "day_of_week": args.day_of_week,
                                "enable": True if args.enable == "true" else False,
                            }
                            r = requests.post(url, json = data)
                            if r.status_code == 200:
                                data = r.json()
                                if raw:
                                    print(json.dumps(data, indent = 4, sort_keys = True))
                                else:
                                    print_table_result(
                                        [data],
                                        ["schedule_id", "result", "message"]
                                    )
                            else:
                                print("error:\ncode: %s\ncontent: %s" % (r.status_code, r.content))
                        except Exception as e:
                            print(e)
                    else:
                        parser.print_help()
                elif operation == "update":
                    if args.schedule_id:
                        try:
                            data = {"schedule_id": args.schedule_id}
                            if args.name is not None:
                                data["schedule_name"] = args.name
                            if args.app_id is not None:
                                data["app_id"] = args.app_id
                            if args.input is not None:
                                data["input_data"] = args.input
                            if args.minute is not None:
                                data["minute"] = args.minute
                            if args.hour is not None:
                                data["hour"] = args.hour
                            if args.day_of_month is not None:
                                data["day_of_month"] = args.day_of_month
                            if args.day_of_week is not None:
                                data["day_of_week"] = args.day_of_week
                            if args.enable is not None:
                                data["enable"] = True if args.enable == "true" else False
                            r = requests.put(url, json = data)
                            if r.status_code == 200:
                                data = r.json()
                                if raw:
                                    print(json.dumps(data, indent = 4, sort_keys = True))
                                else:
                                    print_table_result(
                                        [data],
                                        ["schedule_id", "result", "message"]
                                    )
                            else:
                                print("error:\ncode: %s\ncontent: %s" % (r.status_code, r.content))
                        except Exception as e:
                            print(e)
                    else:
                        parser.print_help()
            else:
                parser.print_help()
        else:
            parser.print_help()
    except Exception as e:
        print(e)


if __name__ == "__main__":
    main()
